#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/******************************************************************
A program to compute the effective conductivity of some composites.

Programmed by Luis Silvestre Feb-2007
*******************************************************************/

#include <iostream>
#include <cstdlib>
#include <cmath>
#include "grid3d.h"
#include "conjgrad.h"
#include "operator3d.h"

#define pi 3.141592653589793238

using namespace std;

double cos3(double x, double y, double z){
  return cos(pi*x)+cos(pi*y)+cos(pi*z);
}

double nodal2(double x, double y, double z){
  return cos(pi*x)+cos(pi*y)+cos(pi*z)
      -0.150763*3*cos(pi*x)*cos(pi*y)*cos(pi*z);
}

double nodal5(double x, double y, double z){
  return 2*(cos(pi*x)+cos(pi*y)+cos(pi*z))
      -0.00956335*6*cos(pi*x)*cos(pi*y)*cos(pi*z)
      +0.00959441*(cos(2*pi*x)*cos(pi*y)+cos(2*pi*x)*cos(pi*z)+cos(2*pi*y)*cos(pi*x)+cos(2*pi*y)*cos(pi*z)+cos(2*pi*z)*cos(pi*x)+cos(2*pi*z)*cos(pi*y))
      -0.0181364*2*(cos(3*pi*x)+cos(3*pi*y)+cos(3*pi*z))
      -0.0292538*2*(cos(2*pi*x)*cos(2*pi*y)*cos(pi*z)+cos(2*pi*x)*cos(pi*y)*cos(2*pi*z)+cos(pi*x)*cos(2*pi*y)*cos(2*pi*z));
}

double midValue(double x, double y, double z){
  if(x>y) swap(x,y);
  if(x>z) swap(x,z);
  if(y>z) swap(y,z);
  return (y-0.5);
}

double optimal(double x, double y, double z){
  return cos(pi*x)+cos(pi*y)+cos(pi*z)-2.25*cos(pi*x)*cos(pi*y)*cos(pi*z);
}

int main(int argc, char *argv[])
{
  cout << "This program computes the effective conductivity of a composite " << endl;
  cout << "made of two phases with conductivities 0.1 and 1.0 respectively." << endl;
  cout << endl << "The value is given by minimizing the functional" << endl;
  cout << "J(u) = \\int_Q a(x) |Du|^2 dx" << endl;
  cout << "from all functions u such that u(x,y,0)=0 and u(x,y,1)=1" << endl;
  cout << "and a(x) takes the values 0.1 and 1.0 in different points inside the cube." << endl;
  cout << endl << "Notice that this imposes Neumann conditions on the side of the cube and" << endl <<" it is equivalent to the symmetric cell problem in the double cube" << endl;
  cout << "*****************************************************************" << endl;

  int n,N,type;
  
  cout << "What type of configuration do you want me to use? " << endl;
  cout << "0 - Just plain Laplace equation with a(x)=1.0" << endl;
  cout << "1 - A checkerboard type configuration with" << endl <<"       a(x) equals 0.1 and 1.0 in adjacent cubes." << endl;
  cout << "2 - A cubical bicontinuous configuration" << endl << "       where a(x) is 0.1 or 1 depending on whether the mid value" << endl << "       of x, y and z is larger or smaller than 1/2" << endl;
  cout << "3 - A bicontinuous configuration with" << endl << "       a(x)=0.1 where cos(pi*x)+cos(pi*y)+cos(pi*z)<0" << endl <<"       a(x)=1.0 where cos(pi*x)+cos(pi*y)+cos(pi*z)>0" << endl;
  cout << "4 - A bicontinuous configuration with a 2 nodal approximation" << endl << "       of the Schwartz P surface." << endl;
  cout << "5 - A bicontinuous configuration with a 5 nodal approximation" << endl << "       of the Schwartz P surface." << endl;
  cout << "6 - A very close to optimal: cos(pi*x)+cos(pi*y)+cos(pi*z)-2.25cos(pi*x)cos(pi*y)cos(pi*z)=0" << endl;
  cin >> type;
  
  if(type>=3){
    cout << endl << "Since the configuration does not match the rectangular grid, an approximate" << endl;
    cout << "configuration is used instead, where a(x) is constant 0.1 or 1.0" << endl<<" in each small cube." << endl;
    cout << "It is done like this not to obtain an artificially large effective conductivity." << endl << endl;
  }
  
  cout << "What grid size do you want? (An odd number is strongly recommended): ";
  cin >> N;
  
  if(N<3){
    cout << "That number is way too small. Using grid size 3 instead." << endl;
    N=3;
  }
  
  n=4;

  grid3d x(n);
  for(int xi=0; xi<n; ++xi){
    for(int yi=0; yi<n; ++yi){
      for(int zi=0; zi<n; ++zi){
        x(xi,yi,zi)=double(zi)/(n-1);
      }
    }
  }
  
  Operator3d* l;
  switch(type){
    case 0: l = new LaplaceOperator;
    break;
    case 1: l = new CheckerBoard;
    break;
    case 2: l = new BicontinuousCubes;
    break;
    default: l = new LevelSet;
  }

  do{
    n = N <? 2*(3*n/4)+3;
    cout << "Init initial guess with grid size " << n << endl;
    x = x.interpolate(n);
    
    if(type==3) ((LevelSet*)l)->init(n,cos3);
    if(type==4) ((LevelSet*)l)->init(n,nodal2);
    if(type==5) ((LevelSet*)l)->init(n,nodal5);
    if(type==6) ((LevelSet*)l)->init(n,optimal);
    grid3d b(l->boundary(n));
    
    cout << "Using conjugate gradients." << endl;
    grid3d s = conjugateGradients(*l,x,b);
    
    cout << endl;
    cout << "********************************************" << endl;
    cout << "End of computation for grid size: " << n << endl;
    cout << "Configuration: ";
    switch(type){
      case 0: cout << "plain Laplacian"; break;
      case 1: cout << "checkers board"; break;
      case 2: cout << "bicontinuous cubes"; break;
      case 3: cout << "one Nodal P"; break;
      case 4: cout << "two Nodal P"; break;
      case 5: cout << "five Nodal P"; break;
      case 6: cout << "bicontinuous far from P"; break;
    }
    cout << endl << "The computed conductivity is " << l->energy(s) << endl;
    cout << "If the value of n is not large enough, the actual conductivity may be smaller." << endl;
    cout <<"These data might be useful: " << endl;
    cout <<"u(1/2,1/2,1/2) - 1/2 = " << s((n-1)/2,(n-1)/2,(n-1)/2) - 0.5 << endl;
    cout <<"          u(1,1/2,0) = " << s(n-1,(n-1)/2,0) << endl;
    cout <<"    u(1,0,1/2) - 1/2 = " << s(n-1,0,(n-1)/2) - 0.5 << endl;
    cout <<"      u(1/2,0,1) - 1 = " << s((n-1)/2,0,n-1) - 1 << endl;
  } while(n<N);
  delete l;
  return EXIT_SUCCESS;
}
