#include "operator3d.h"
#include <iostream>

inline double sqr(double x){
  return x*x;
}

grid3d Operator3d::operator()(const grid3d& g) const{
  int n = g.dimension();
  grid3d r(n);
  for(int x=0; x<n; ++x){
    for(int y=0; y<n; ++y){
      for(int z=1; z<n-1; ++z){
        coefs c = findCoefs(x,y,z,n);
        r(x,y,z) = (c.x1+c.x2+c.y1+c.y2+c.z1+c.z2)*g.get(x,y,z) 
            -c.x1*((x==0) ? g.get(1,y,z) : g.get(x-1,y,z))-c.x2*((x==n-1)? g.get(n-2,y,z) : g.get(x+1,y,z))
            -c.y1*((y==0) ? g.get(x,1,z) : g.get(x,y-1,z))-c.y2*((y==n-1)? g.get(x,n-2,z) : g.get(x,y+1,z))
            -c.z1*((z==1) ? 0 : g.get(x,y,z-1))-c.z2*((z==n-2) ? 0 : g.get(x,y,z+1));
      }
    }
  }
  for(int x=0; x<n; ++x){
    for(int y=0; y<n; ++y){
      r(x,y,0) = g.get(x,y,0);
      r(x,y,n-1) = g.get(x,y,n-1);
    }
  }
  return r;
}

grid3d Operator3d::boundary(int n) const{
  grid3d r(n);
  r.clear();
  for(int x=0; x<n; ++x){
    for(int y=0; y<n; ++y){
      coefs c = findCoefs(x,y,n-2,n);
      r(x,y,n-1) = 1;
      r(x,y,n-2) = c.z2;
    }
  }
  return r;
}

double Operator3d::energy(const grid3d& g) const{
  int n = g.dimension();
  double h = 1.0/(n-1);
  double r=0;
  for(int x=0; x<n; ++x){
    for(int y=0; y<n; ++y){
      for(int z=0; z<n; ++z){
        coefs c = findCoefs(x,y,z,n);
        double t;
        t = c.x1*sqr(g.get(x,y,z) - ((x==0) ? g.get(1,y,z) : g.get(x-1,y,z)))
            +c.x2*sqr(g.get(x,y,z) - ((x==n-1)? g.get(n-2,y,z) : g.get(x+1,y,z)))
            +c.y1*sqr(g.get(x,y,z) - ((y==0) ? g.get(x,1,z) : g.get(x,y-1,z)))
            +c.y2*sqr(g.get(x,y,z) - ((y==n-1)? g.get(x,n-2,z) : g.get(x,y+1,z)))
            +c.z1*sqr(g.get(x,y,z) - ((z==0) ? g.get(x,y,1) : g.get(x,y,z-1)))
            +c.z2*sqr(g.get(x,y,z) - ((z==n-1) ? g.get(x,y,n-2) : g.get(x,y,z+1)));
        if ((x==0)||(x==n-1)) t /= 2;
        if ((y==0)||(y==n-1)) t /= 2;
        if ((z==0)||(z==n-1)) t /= 2;
        r += t;
      }
    }
  }
  return r/(2*sqr(h)*(n-1)*(n-1)*(n-1));
}


coefs LaplaceOperator::findCoefs(int x, int y, int z, int n) const{
  return (coefs){1.0,1.0,1.0,1.0,1.0,1.0};
}

coefs BicontinuousCubes::findCoefs(int x, int y, int z, int n) const{
  int med = (n-1)/2;
  
  // General points
  if( ((x>med)&&(y>med)) || ((x>med)&&(z>med)) || ((z>med)&&(y>med)) ) 
    return (coefs){s1,s1,s1,s1,s1,s1};
  if( ((x<med)&&(y<med)) || ((x<med)&&(z<med)) || ((z<med)&&(y<med)) ) 
    return (coefs){s2,s2,s2,s2,s2,s2};
  
  // On the sides
  if((x==med)&&(y<med)&&(z>med)) return (coefs){s2,s1,(s1+s2)/2,(s1+s2)/2,(s1+s2)/2,(s1+s2)/2};
  if((x==med)&&(y>med)&&(z<med)) return (coefs){s2,s1,(s1+s2)/2,(s1+s2)/2,(s1+s2)/2,(s1+s2)/2};
  if((x<med)&&(y==med)&&(z>med)) return (coefs){(s1+s2)/2,(s1+s2)/2,s2,s1,(s1+s2)/2,(s1+s2)/2};
  if((x>med)&&(y==med)&&(z<med)) return (coefs){(s1+s2)/2,(s1+s2)/2,s2,s1,(s1+s2)/2,(s1+s2)/2};
  if((x<med)&&(y>med)&&(z==med)) return (coefs){(s1+s2)/2,(s1+s2)/2,(s1+s2)/2,(s1+s2)/2,s2,s1};
  if((x>med)&&(y<med)&&(z==med)) return (coefs){(s1+s2)/2,(s1+s2)/2,(s1+s2)/2,(s1+s2)/2,s2,s1};
  
  // On the edges
  if((x==med)&&(y==med)&&(z<med)) return (coefs){s2,(s1+s2)/2,s2,(s1+s2)/2,(s1+3*s2)/4,(s1+3*s2)/4};
  if((x==med)&&(y==med)&&(z>med)) return (coefs){(s1+s2)/2,s1,(s1+s2)/2,s1,(3*s1+s2)/4,(3*s1+s2)/4};
  if((x==med)&&(y<med)&&(z==med)) return (coefs){s2,(s1+s2)/2,(s1+3*s2)/4,(s1+3*s2)/4,s2,(s1+s2)/2};
  if((x==med)&&(y>med)&&(z==med)) return (coefs){(s1+s2)/2,s1,(3*s1+s2)/4,(3*s1+s2)/4,(s1+s2)/2,s1};
  if((x<med)&&(y==med)&&(z==med)) return (coefs){(s1+3*s2)/4,(s1+3*s2)/4,s2,(s1+s2)/2,s2,(s1+s2)/2};
  if((x>med)&&(y==med)&&(z==med)) return (coefs){(3*s1+s2)/4,(3*s1+s2)/4,(s1+s2)/2,s1,(s1+s2)/2,s1};
  
  // On the vertex
  if((x==med)&&(y==med)&&(z==med)) return (coefs){(s1+3*s2)/4,(3*s1+s2)/4,(s1+3*s2)/4,(3*s1+s2)/4,(s1+3*s2)/4,(3*s1+s2)/4};

  // The flow should never reach this point
  cerr << "There is a bug in BicontinuousCubes::findCoefs, with arguments: " << x << ", " << y << ", " << z << ", " << n << endl;
  return (coefs) {-10,-10,-10,-10,-10,-10};
}

coefs CheckerBoard::findCoefs(int x, int y, int z, int n) const{
  int med = (n-1)/2;
  
  // General points
  if( ((x>med)&&(y>med)&&(z>med)) || ((x>med)&&(y<med)&&(z<med)) || ((x<med)&&(y>med)&&(z<med)) || ((x<med)&&(y<med)&&(z>med)) )
    return (coefs){s1,s1,s1,s1,s1,s1};
  if( ((x>med)&&(y>med)&&(z<med)) || ((x>med)&&(y<med)&&(z>med)) || ((x<med)&&(y>med)&&(z>med)) || ((x<med)&&(y<med)&&(z<med)) )
    return (coefs){s2,s2,s2,s2,s2,s2};
  
  // On the sides
  if((x==med)&&(y<med)&&(z<med)) return (coefs){s2,s1,(s1+s2)/2,(s1+s2)/2,(s1+s2)/2,(s1+s2)/2};
  if((x==med)&&(y>med)&&(z>med)) return (coefs){s2,s1,(s1+s2)/2,(s1+s2)/2,(s1+s2)/2,(s1+s2)/2};
  if((x==med)&&(y<med)&&(z>med)) return (coefs){s1,s2,(s1+s2)/2,(s1+s2)/2,(s1+s2)/2,(s1+s2)/2};
  if((x==med)&&(y>med)&&(z<med)) return (coefs){s1,s2,(s1+s2)/2,(s1+s2)/2,(s1+s2)/2,(s1+s2)/2};
  
  if((x<med)&&(y==med)&&(z<med)) return (coefs){(s1+s2)/2,(s1+s2)/2,s2,s1,(s1+s2)/2,(s1+s2)/2};
  if((x>med)&&(y==med)&&(z>med)) return (coefs){(s1+s2)/2,(s1+s2)/2,s2,s1,(s1+s2)/2,(s1+s2)/2};
  if((x<med)&&(y==med)&&(z>med)) return (coefs){(s1+s2)/2,(s1+s2)/2,s1,s2,(s1+s2)/2,(s1+s2)/2};
  if((x>med)&&(y==med)&&(z<med)) return (coefs){(s1+s2)/2,(s1+s2)/2,s1,s2,(s1+s2)/2,(s1+s2)/2};
  
  if((x<med)&&(y<med)&&(z==med)) return (coefs){(s1+s2)/2,(s1+s2)/2,(s1+s2)/2,(s1+s2)/2,s2,s1};
  if((x>med)&&(y>med)&&(z==med)) return (coefs){(s1+s2)/2,(s1+s2)/2,(s1+s2)/2,(s1+s2)/2,s2,s1};
  if((x<med)&&(y>med)&&(z==med)) return (coefs){(s1+s2)/2,(s1+s2)/2,(s1+s2)/2,(s1+s2)/2,s1,s2};
  if((x>med)&&(y<med)&&(z==med)) return (coefs){(s1+s2)/2,(s1+s2)/2,(s1+s2)/2,(s1+s2)/2,s1,s2};
  
  // Amazingly every other case gives the same coeffs
  return (coefs){(s1+s2)/2,(s1+s2)/2,(s1+s2)/2,(s1+s2)/2,(s1+s2)/2,(s1+s2)/2};
}

coefs LevelSet::findCoefs(int x, int y, int z, int n) const{
  coefs r;
  if(n != _n) cerr << "LevelSet::findCoefs was called for a different n that what it was initialized for." << endl;
  int x1 = (x==0) ? 0 : x-1;
  int y1 = (y==0) ? 0 : y-1;
  int z1 = (z==0) ? 0 : z-1;
  r.x1 = (c->get(x1,y,z) + c->get(x1,y1,z) + c->get(x1,y,z1) + c->get(x1,y1,z1)) / 4;
  r.x2 = (c->get(x,y,z) + c->get(x,y1,z) + c->get(x,y,z1) + c->get(x,y1,z1)) / 4;
  r.y1 = (c->get(x,y1,z) + c->get(x1,y1,z) + c->get(x,y1,z1) + c->get(x1,y1,z1)) / 4;
  r.y2 = (c->get(x,y,z) + c->get(x1,y,z) + c->get(x,y,z1) + c->get(x1,y,z1)) / 4;
  r.z1 = (c->get(x,y,z1) + c->get(x1,y,z1) + c->get(x,y1,z1) + c->get(x1,y1,z1)) / 4;
  r.z2 = (c->get(x,y,z) + c->get(x1,y,z) + c->get(x,y1,z) + c->get(x1,y1,z)) / 4;
  return r;
}

LevelSet::LevelSet() : Operator3d() {
  _n = 0;
  c = NULL;
}

LevelSet::~LevelSet(){
  if (c!=NULL) delete c;
}

void LevelSet::init(int n, double f(double, double, double)){
  if (c != NULL) delete c;
  _n = n;
  c = new grid3d(n);
  int c1=0,c2=0,c3=0;
  for(int x=0; x<n; ++x){
    for(int y=0; y<n; ++y){
      for(int z=0; z<n; ++z){
        double x0 = (double(x)+0.5)/(n-1);
        double y0 = (double(y)+0.5)/(n-1);
        double z0 = (double(z)+0.5)/(n-1);
        if( f(x0,y0,z0) > 0 ){
          c->at(x,y,z) = 1;
          if((x!=n-1) && (y!=n-1) && (z!=n-1)) ++c1;
        }
        else{ 
          if( f(x0,y0,z0) < 0 ){
            c->at(x,y,z) = s2; 
            if((x!=n-1) && (y!=n-1) && (z!=n-1)) ++c2;
          }
          else {
            c->at(x,y,z) = 0; // I hope this never happens. But I pull it down anyway.
            ++c3;
          }
        }
      }
    }
  }
  if(c1 != c2) cerr << "Warning. The coefficients s2 and s1 are not evenly distributed in the unit cube." << endl;
  if(c3 > 0) cerr << "Warning. Some cube centers hit the level set exactly. Setting those cubes as 0 just in case." << endl;
}
