#ifndef operator3d
#define operator3d

/*************************************************
Programmed by Luis Silvestre Feb-2007
**************************************************/

#include "grid3d.h"

// The conductivities:
#define s1 1.0
#define s2 0.1

struct coefs{
  double x1,x2,y1,y2,z1,z2;
};

/*************************************************************
Operator3d is an abstract class to be used to solve the cell
problem. Each subclass defines a different configuration.
*************************************************************/
class Operator3d{
  protected:
    virtual coefs findCoefs(int x, int y, int z, int n) const=0;
  public:
    grid3d boundary(int n) const;
    grid3d operator()(const grid3d& g) const;
    double energy(const grid3d& g) const;
};

/*************************************************************
LaplaceOperator is only for testing purposes
It should solve the Laplace equation with boundary data
0 on (x,y,0), 1 on (x,y,1) and Neumann everywhere else.
It should always return 1 as the eff. conductivity.
*************************************************************/
class LaplaceOperator : public Operator3d{
  protected:
    coefs findCoefs(int x, int y, int z, int n) const;
};

/*************************************************************
BicontinuousCubes is for conductivity:
a(x,y,z) = 1   if the middle coordinate is > 1/2
a(x,y,z) = 0.1 if the middle coordinate is < 1/2
where the middle coordinate refers to the value of x, y or z 
that is neither the maximum nor the minimum.
The effective conductivity gives 0.419464 w/ 101 grid points.
(Recall the max from all config. is 0.427)
*************************************************************/
class BicontinuousCubes : public Operator3d{
  protected:
    coefs findCoefs(int x, int y, int z, int n) const;
};

/*************************************************************
CheckerBoard is for conductivity:
a(x,y,z) = 1   if the [2x+2y+2z] is even
a(x,y,z) = 0.1 if the [2x+2y+2z] is odd
where [t] refers to the integer part of t
*************************************************************/
class CheckerBoard : public Operator3d{
  protected:
    coefs findCoefs(int x, int y, int z, int n) const;
};

/*************************************************************
In LevelSet, the two phases are given by
a(x,y,z) = 1   if f(x,y,z)>0
a(x,y,z) = 0.1 if f(x,y,z)<0
where f must be given previously with LevelSet::init()
*************************************************************/
class LevelSet  : public Operator3d{
  private:
  // Stores in c the coefficient for each cube (that will be either s1 or s2)
    grid3d* c;
    int _n;
  protected:
    coefs findCoefs(int x, int y, int z, int n) const;
  public: 
    LevelSet();
    virtual ~LevelSet();
    // This must be called first to initialize the variable c
    // n is the grid size used. The two phases will be divided by f=0
    void init(int n, double f(double, double, double));
};

#endif
